/*
  This file is part of GNOME Taquin

  GNOME Taquin is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  GNOME Taquin is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with GNOME Taquin.  If not, see <https://www.gnu.org/licenses/>.
*/

@import url("base-window.css");

button.unfullscreen-button {
  margin:6px;
}

/*\
* * board generics; TODO move in game-window.css
\*/

                               .game-box { transition:padding 0.3s; padding:1.5rem; }
                  .thin-window .game-box {                          padding:1.0rem; }
.extra-thin-window.thin-window .game-box,
                  .flat-window .game-box {                          padding:0.5rem; }
.extra-flat-window.flat-window .game-box {                          padding:0.4rem; }

/*\
* * start-game button; TODO move in game-window.css
\*/

.extra-flat-window             button.start-game-button {
  min-height:2rem;
}
                               button.start-game-button {
  min-height:3rem;
  min-width:11rem;

  transition:min-height          0.3s ease 0.01s;
}

                               button.start-game-button { margin-top:1.5rem; margin-bottom:0.5rem;
                                               transition:margin-top 0 ease, margin-bottom 0 ease; }
                  .flat-window button.start-game-button { margin-top:0.5rem;                       }
.extra-flat-window.flat-window button.start-game-button { margin-top:  0rem; margin-bottom:0.4rem; }

.extra-thin-window             button.start-game-button { margin-top:1.0rem; }

/* TODO move in game-window.css */
label.game-name-label:dir(ltr) {
  margin-left : 12px;
}
label.game-name-label:dir(rtl) {
  margin-right: 12px;
}

/* the actionbar is in an overlay... */
actionbar {
  background-color:@theme_bg_color;
}
actionbar:backdrop {
  background-color:@theme_unfocused_bg_color;
}

/*\
* * games buttons
\*/

button.radio-like-toggle#radio-one { background-image: url("resource://org/gnome/Taquin/images/15-Puzzle.png"); }
button.radio-like-toggle#radio-two { background-image: url("resource://org/gnome/Taquin/images/16-Puzzle.png"); }

/*\
* * the end
\*/
